# pylint: disable=no-member,no-classmethod-decorator,protected-access
from __future__ import unicode_literals
from functools import lru_cache
from django_filters import rest_framework as filters, filterset
from django.db.models.base import ModelBase, Model
from ..utils import import_class, apply_decorators, classproperty
from ..api import (
    base as api_base,
    filters as api_filters,
    serializers as api_serializers,
    decorators as api_decorators
)
from .queryset import BQuerySet
from .manager import BaseManager


def _get_setting_for_view(metatype, metadata, views):
    override = metadata[f'override_{metatype}']
    metadataobject = metadata[metatype]
    if override:
        return metadataobject  # nocv
    if metadataobject:
        for view in views:
            if hasattr(view, metatype):
                return list(getattr(view, metatype)) + list(metadata[metatype])
        return metadataobject  # nocv


class ApplyNestedDecorators(apply_decorators):
    def __init__(self, nested: dict):
        super().__init__(
            *map(self.__get_decorator, nested.items())
        )

    def __get_decorator(self, data):
        path, deco_kwargs = data

        assert 'model' in deco_kwargs and 'view' not in deco_kwargs, (
            "Invalid model configuration: "
            f"Unable to set 'model' and 'view' at the same time for path [{path}]."
        )
        if 'model' in deco_kwargs:
            model = deco_kwargs.pop('model')
            assert isinstance(model, ModelBaseClass), (
                f"Invalid model type {type(model)} for path [{path}]."
            )
            deco_kwargs['view'] = model.generated_view
            if 'arg' not in deco_kwargs:
                deco_kwargs['arg'] = model._meta.pk.name
        return api_decorators.nested_view(path, **deco_kwargs)


class ModelBaseClass(ModelBase):
    """Metaclass for all models."""

    def __new__(mcs, name, bases, attrs, **kwargs):
        if "__slots__" not in attrs:
            attrs['__slots__'] = tuple()
        if "__unicode__" in attrs and '__str__' not in attrs:
            attrs['__str__'] = lambda x: x.__unicode__()
        extra_metadata: dict = {
            # list or class which is base for view
            "view_class": None,
            # base class for serializers
            "serializer_class": None,
            # name of openapi model
            "serializer_class_name": None,
            # tuple or list of fields in list view
            "list_fields": None,
            # dict which override fields types of list view serializer
            "override_list_fields": dict(),
            # tuple or list of fields in detail view
            "detail_fields": None,
            # dict which override fields types of detail view serializer
            "override_detail_fields": dict(),
            # key-value of actions serializers (key - action, value - serializer class)
            "extra_serializer_classes": dict(),
            # tuple or list of filters on list
            "filterset_fields": 'serializer',
            # tuple or list of filter backends for queryset
            "filter_backends": None,
            # allow to full override of the filter backends default list
            "override_filter_backends": False,
            # tuple or list of permission_classes for the view
            "permission_classes": None,
            # allow to override the default permission_classes
            "override_permission_classes": False,
            # additional attrs which means that this view allowed to copy elements
            "copy_attrs": dict(),
            # key-value mapping with nested views (key - nested name, kwargs for nested decorator)
            "nested": {}
        }
        if "Meta" in attrs:
            meta = attrs['Meta'].__dict__
            for extra_name in filter(lambda y: y in meta, map(lambda x: f'_{x}', extra_metadata.keys())):
                extra_metadata[extra_name[1:]] = meta[extra_name]
        attrs['__extra_metadata__'] = extra_metadata
        model_class: Model = super(ModelBaseClass, mcs).__new__(mcs, name, bases, attrs, **kwargs)
        queryset_class = getattr(getattr(model_class, 'objects', None), '_queryset_class', None)
        if queryset_class and not issubclass(queryset_class, BQuerySet):
            manager = BaseManager.from_queryset(BQuerySet)()
            manager.auto_created = True
            model_class.add_to_class('objects', manager)
        if hasattr(model_class, '__prepare_model__'):
            model_class.__prepare_model__()
        return model_class

    @classproperty
    @lru_cache()
    def generated_view(cls):
        # pylint: disable=no-value-for-parameter
        return cls.get_view_class()

    def get_serializer_class(cls, serializer_class, serializer_class_name=None, fields=None, field_overrides=None):
        if serializer_class is None:
            serializer_class = api_serializers.VSTSerializer

        if serializer_class_name is None:
            serializer_class_name = cls.__name__ + 'Serializer'

        if fields:
            fields = list(fields)
        else:
            fields = '__all__'

        class SerializerClass(serializer_class):
            class Meta:
                model = cls
                ref_name = serializer_class_name.replace('Serializer', '')

        for attr_name, attr_value in field_overrides.items():
            SerializerClass._declared_fields[attr_name] = attr_value

        SerializerClass.__name__ = serializer_class_name
        setattr(SerializerClass.Meta, 'fields', fields)

        return SerializerClass

    def get_extra_metadata(cls):
        return cls.__extra_metadata__

    def get_view_class(cls):
        metadata = cls.get_extra_metadata()  # pylint: disable=no-value-for-parameter

        view_attributes = dict(model=cls)

        serializer_class = metadata['serializer_class']
        serializers = dict(
            serializer_class=cls.get_serializer_class(  # pylint: disable=no-value-for-parameter
                serializer_class=serializer_class,
                serializer_class_name=metadata['serializer_class_name'],
                fields=metadata['list_fields'],
                field_overrides=metadata['override_list_fields']
            )
        )
        detail_fields_override = metadata['override_detail_fields'] or None
        if not detail_fields_override and not metadata['detail_fields']:
            detail_fields_override = metadata['override_list_fields']

        serializers['serializer_class_one'] = cls.get_serializer_class(  # pylint: disable=no-value-for-parameter
            serializer_class=serializer_class,
            serializer_class_name=f'One{serializers["serializer_class"].__name__}',
            fields=metadata['detail_fields'] or metadata['list_fields'],
            field_overrides=detail_fields_override
        )
        serializers.update(map(lambda k, v: (f'serializer_class_{k}', v), metadata['extra_serializer_classes'].items()))

        view_class = metadata['view_class']
        if view_class is None:
            view_class = api_base.ModelViewSet
        if view_class == 'read_only':
            view_class = api_base.ReadOnlyModelViewSet
        elif view_class == 'history':
            view_class = api_base.HistoryModelViewSet  # nocv
        elif isinstance(view_class, str):
            view_class = import_class(view_class)

        if not isinstance(view_class, (tuple, list)):
            view_class = (view_class,)

        view_class = list(view_class)

        if metadata['copy_attrs']:
            view_attributes.update(map(lambda r: (f'copy_{r[0]}', r[1]), metadata['copy_attrs'].items()))
            view_class.append(api_base.CopyMixin)

        filterset_fields = metadata['filterset_fields']
        if filterset_fields == 'serializer':
            filterset_fields = tuple(serializers['serializer_class']._declared_fields.keys())

        if filterset_fields:

            if isinstance(filterset_fields, dict):
                filterset_fields_list = tuple(filterset_fields.keys())
                filterset_fields_types = filterset_fields
            else:
                filterset_fields_list = filterset_fields
                filterset_fields_types = dict()

            class Meta:
                model = cls
                fields = filterset_fields_list

            view_attributes['filterset_class'] = filterset.FilterSetMetaclass(
                'FilterSetClass',
                (api_filters.DefaultIDFilter if 'id' in filterset_fields_list else filters.FilterSet,),
                {'Meta': Meta, **filterset_fields_types}
            )

        for metatype in ['permission_classes', 'filter_backends']:
            metaobject = _get_setting_for_view(metatype, metadata, view_class)
            if metaobject:
                view_attributes[metatype] = metaobject

        generated_view = type(
            f'{cls.__name__}ViewSet',
            tuple(view_class),
            {
                **view_attributes,
                **serializers,
                **dict(filter(lambda x: hasattr(x[1], '_append_to_view'), vars(cls).items()))
            }
        )

        return apply_decorators(
            *getattr(cls, 'generated_view_decorators', [])
        )(ApplyNestedDecorators(metadata['nested'])(generated_view))



#  pylint: disable=invalid-name
"""
WSGI config for VST Projects project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/howto/deployment/wsgi/
"""

from django.core.wsgi import get_wsgi_application
from .environment import prepare_environment

prepare_environment()

application = get_wsgi_application()



import inspect
from django.db import models


def is_class_method_or_function(obj):
    return inspect.isfunction(obj) or \
           inspect.ismethod(obj) or \
           isinstance(obj, type(is_class_method_or_function))


class BaseManager(models.Manager):

    @classmethod
    def _get_queryset_methods(cls, queryset_class):
        """
        Django overrloaded method for add cyfunction.
        """
        def create_method(name, method):  # nocv
            def manager_method(self, *args, **kwargs):
                return getattr(self.get_queryset(), name)(*args, **kwargs)

            manager_method.__name__ = method.__name__
            manager_method.__doc__ = method.__doc__
            return manager_method

        orig_method = models.Manager._get_queryset_methods
        new_methods = orig_method(queryset_class)
        for name, method in inspect.getmembers(queryset_class, predicate=is_class_method_or_function):
            # Only copy missing methods.
            if hasattr(cls, name) or name in new_methods:
                continue
            queryset_only = getattr(method, 'queryset_only', None)
            if queryset_only or (queryset_only is None and name.startswith('_')):
                continue
            # Copy the method onto the manager.
            new_methods[name] = create_method(name, method)  # nocv
        return new_methods



from django_filters import rest_framework as filters
from django_filters import CharFilter

id_help = 'A unique integer value (or comma separated list) identifying this instance.'
name_help = 'A name string value (or comma separated list) of instance.'


def _extra_search(queryset, field, value, stype):
    vals = field.split("__")
    field, tp = vals[0], (list(vals)[1:2] + [""])[0]
    field += f"__{stype}"
    value = value.split(",") if stype == "in" else value
    if tp.upper() == "NOT":
        return queryset.exclude(**{field: value})
    return queryset.filter(**{field: value})


def extra_filter(queryset, field, value):
    return _extra_search(queryset, field, value, "in")


def name_filter(queryset, field, value):
    return _extra_search(queryset, field, value, "contains")


class DefaultIDFilter(filters.FilterSet):
    id = CharFilter(method=extra_filter, help_text=id_help)
    id__not = CharFilter(method=extra_filter, help_text=id_help)



"""
Default Django model classes overrides in `vstutils.models` module.
"""

from __future__ import unicode_literals
from django.db import models

from .base import ModelBaseClass
from .manager import BaseManager
from .queryset import BQuerySet
from .model import BaseModel
from .decorators import register_view_action, register_view_method


class Manager(BaseManager.from_queryset(BQuerySet)):  # type: ignore
    """
    Default VSTUtils manager. Used by `BaseModel` and `BModel`.
    Allows to use managers and querysets with cyfunctions-members.
    """


class BModel(BaseModel):
    """
    Default model class with usefull attributes.

    Examples:
        .. sourcecode:: python

            from django.db import models
            from rest_framework.fields import ChoiceField
            from vstutils.models import BModel

            class Stage(BModel):
                name = models.CharField(max_length=256)
                order = models.IntegerField(default=0)

                class Meta:
                    default_related_name = "stage"
                    ordering = ('order', 'id',)
                    # fields which would be showed on list.
                    _list_fields = [
                        'id',
                        'name',
                    ]
                    # fields which would be showed on detail view and creation.
                    _detail_fields = [
                        'id',
                        'name',
                        'order'
                    ]
                    # make order as choices from 0 to 9
                    _override_detail_fields = {
                        'order': ChoiceField((str(i) for i in range(10)))
                    }



            class Task(BModel):
                name = models.CharField(max_length=256)
                stages = models.ManyToManyField(Stage)

                class Meta:
                    # fields which would be showed.
                    _list_fields = [
                        'id',
                        'name',
                    ]
                    # create nested views from models
                    _nested = {
                        'stage': {
                            'allow_append': False,
                            'model': Stage
                        }
                    }


        In this case, you create models which could converted to simple view, where:

        - ``POST``/``GET`` to ``/api/version/task/`` - creates new or get list of tasks
        - ``PUT``/``PATCH``/``GET``/``DELETE`` to ``/api/version/task/:id/`` - updates, retrieves or removes instance of task
        - ``POST``/``GET` to ``/api/version/task/:id/stage/`` - creates new or get list of stages in task
        - ``PUT``/``PATCH``/``GET``/``DELETE`` to ``/api/version/task/:id/stage/:stage_id`` - updates, retrieves or removes instance of stage in task.
    """

    #: Primary field for select and search in API.
    id = models.AutoField(primary_key=True, max_length=20)
    #: Useful field for hidden data.
    hidden = models.BooleanField(default=False)

    class Meta:
        abstract = True

    def __unicode__(self):
        return f"<{self.id}>"
